#include <torch/extension.h>
#include <ATen/ATen.h>
#include <vector>

#define CHECK_CUDA(x) TORCH_CHECK(x.is_cuda(), #x " must be a CUDA tensor")
#define CHECK_CONTIGUOUS(x) TORCH_CHECK(x.is_contiguous(), #x " must be contiguous")
#define CHECK_INPUT(x) CHECK_CUDA(x); CHECK_CONTIGUOUS(x)

// Tile dimension for shared memory tiling
#define TILE_DIM 16

// Optimized assignment computation kernel using shared memory tiling
__global__ void optimizedTiledAssignment(const float* __restrict__ A, const float* __restrict__ B, float* __restrict__ C,
                              int M, int D, int KplusG) {
  int row = blockIdx.y * TILE_DIM + threadIdx.y;
  int col = blockIdx.x * TILE_DIM + threadIdx.x;

  __shared__ float tile_A[TILE_DIM][TILE_DIM];
  __shared__ float tile_B[TILE_DIM][TILE_DIM];

  float sum = 0.0f;

  for (int t = 0; t < (D + TILE_DIM - 1) / TILE_DIM; t++) {
    int a_col = t * TILE_DIM + threadIdx.x;
    if (row < M && a_col < D) {
      tile_A[threadIdx.y][threadIdx.x] = A[row * D + a_col];
    } else {
      tile_A[threadIdx.y][threadIdx.x] = 0.0f;
    }

    int b_row = t * TILE_DIM + threadIdx.y;
    if (b_row < D && col < KplusG) {
      tile_B[threadIdx.y][threadIdx.x] = B[b_row * KplusG + col];
    } else {
      tile_B[threadIdx.y][threadIdx.x] = 0.0f;
    }

    __syncthreads();

    #pragma unroll
    for (int i = 0; i < TILE_DIM; i++) {
      sum += tile_A[threadIdx.y][i] * tile_B[i][threadIdx.x];
    }
    __syncthreads();
  }

  if (row < M && col < KplusG) {
    C[row * KplusG + col] = sum;
  }
}

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor clusters,
    torch::Tensor clusters2,
    torch::Tensor bn_weight,
    torch::Tensor bn_bias,
    torch::Tensor bn_running_mean,
    torch::Tensor bn_running_var,
    int64_t feature_size,
    int64_t cluster_size,
    bool is_training
) {
    CHECK_INPUT(x);
    CHECK_INPUT(clusters);
    CHECK_INPUT(clusters2);
    CHECK_INPUT(bn_weight);
    CHECK_INPUT(bn_bias);
    CHECK_INPUT(bn_running_mean);
    CHECK_INPUT(bn_running_var);

    int64_t B = x.size(0);
    int64_t N = x.size(1);
    int64_t D = feature_size;
    int64_t K = cluster_size;
    int64_t KplusG = clusters.size(1);

    auto x_reshaped = x.reshape({B * N, D});

    if (x_reshaped.device() != clusters.device()) {
        TORCH_CHECK(false, "x.device() != clusters.device()");
    }

    auto assignment = torch::empty({B * N, KplusG}, x.options());

    int M = B * N;
    dim3 block(TILE_DIM, TILE_DIM);
    dim3 grid((KplusG + TILE_DIM - 1) / TILE_DIM, (M + TILE_DIM - 1) / TILE_DIM);

    optimizedTiledAssignment<<<grid, block>>>(x_reshaped.data_ptr<float>(), clusters.data_ptr<float>(), assignment.data_ptr<float>(), M, D, KplusG);

    assignment = torch::batch_norm(
        assignment,
        bn_weight,
        bn_bias,
        bn_running_mean,
        bn_running_var,
        is_training,
        0.1,
        1e-5,
        true
    );

    assignment = torch::softmax(assignment, 1);

    assignment = assignment.narrow(1, 0, K);

    assignment = assignment.reshape({B, N, K});

    auto a_sum = assignment.sum(1, /*keepdim=*/true);

    auto clusters2_exp = clusters2.expand({B, D, K});

    auto a = clusters2_exp * a_sum;

    assignment = assignment.transpose(1, 2);

    auto x_orig = x.reshape({B, N, D});

    auto vlad = torch::bmm(assignment, x_orig);

    vlad = vlad.transpose(1, 2);

    vlad = vlad - a;

    vlad = torch::nn::functional::normalize(
        vlad, torch::nn::functional::NormalizeFuncOptions().p(2).dim(1));

    vlad = vlad.reshape({B, D * K});

    vlad = torch::nn::functional::normalize(
        vlad, torch::nn::functional::NormalizeFuncOptions().p(2).dim(1));

    return vlad;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "NetVLAD forward with optimized assignment computation");
}
